from pydantic import BaseModel
from typing import Any, Dict, List, Optional

class DataCreate(BaseModel):
    """
    Model for creating data.
    
    Attributes
    ----------
    title : str
        Title of the dataset stored in metadata.
    description : str
        Description of dataset stored in metadata.
    source : str
        Data source stored in metadata.
    data : list(dict)
        The data itself for the dataset. Should be a list of dictionaries with the same keys, where each key in each dict is a column name.
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from msdss_data_api.models import *
        from pprint import pprint

        fields = DataCreate.__fields__
        pprint(fields)
    """
    title: Optional[str]
    description: Optional[str]
    source: Optional[str]
    data: List[Dict[str, Any]]

class MetadataUpdate(BaseModel):
    """
    Model for updating metadata.
    
    Attributes
    ----------
    title : str
        Title of the dataset stored in metadata.
    description : str
        Description of dataset stored in metadata.
    source : str
        Data source stored in metadata.
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from msdss_data_api.models import *
        from pprint import pprint

        fields = MetadataUpdate.__fields__
        pprint(fields)
    """
    title: Optional[str]
    description: Optional[str]
    source: Optional[str]
    


from fastapi import HTTPException
from msdss_base_database import Database
from msdss_base_database.defaults import DEFAULT_SUPPORTED_OPERATORS

from .defaults import *

class DataHandler:
    """
    Class to handle dataset events.
    
    Parameters
    ----------
    database : :class:`msdss_base_database:msdss_base_database.core.Database` or None
        Database object to use for managing datasets. If ``None``, a default database will be used.
    permitted_tables : list(str)
        List of permitted table names that are only accessible. If any tables not in this list are accessed, a 401 unauthorized http exception will be thrown.
    restricted_tables : list(str)
        List of restricted table names that are not accessible. If any of these are accessed, a 401 unauthorized http exception will be thrown.
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from msdss_base_database import Database
        from msdss_data_api.handlers import *
        from msdss_data_api.managers import *
        
        # Setup objects
        db = Database()
        dm = DataManager(database=db)
        handler = DataHandler(database=db)

        # Check if the table exists and drop if it does
        if db.has_table('test_table'):
            db.drop_table('test_table')

        # Check table before writing
        # Should not raise exceptions
        handler.handle_write('test_table')

        # Create sample data
        data = [
            {'id': 1, 'column_one': 'a', 'column_two': 2},
            {'id': 2, 'column_one': 'b', 'column_two': 4},
            {'id': 3, 'column_one': 'c', 'column_two': 6},
        ]
        dm.create('test_table', data)

        # Check table name
        # Should not raise exceptions
        handler.handle_read('test_table')

        # Check table restrictions
        # Should not raise exceptions
        handler.handle_restrictions('test_table')
    """
    def __init__(self, database=None, permitted_tables=[], restricted_tables=DEFAULT_RESTRICTED_TABLES):
        database = database if database else Database()
        self.database = database
        self.permitted_tables = permitted_tables
        self.restricted_tables = restricted_tables

    def handle_delete(self, dataset, where_list, delete_all):
        """
        Handle a table delete operation.
        
        Parameters
        ----------
        dataset : str
            Name of the table to check.
        where_list : list(list(str))
            list of where statements the form of ``['column', 'operator', 'value']`` to further filter individual values or rows.
            
            * Operators are one of:

                .. jupyter-execute::
                    :hide-code:

                    from msdss_base_database.defaults import DEFAULT_SUPPORTED_OPERATORS
                    for operator in DEFAULT_SUPPORTED_OPERATORS:
                        print(operator)

            * Example: ``['column_two', '<'. '3']'``

        delete_all : bool
            Whether to delete the entire table or not.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            from msdss_base_database import Database
            from msdss_data_api.handlers import *
            from msdss_data_api.managers import *
            
            # Setup objects
            db = Database()
            dm = DataManager(database=db)
            handler = DataHandler(database=db)

            # Check if the table exists and drop if it does
            if db.has_table('test_table'):
                db.drop_table('test_table')

            # Create sample data
            data = [
                {'id': 1, 'column_one': 'a', 'column_two': 2},
                {'id': 2, 'column_one': 'b', 'column_two': 4},
                {'id': 3, 'column_one': 'c', 'column_two': 6},
            ]
            dm.create('test_table', data)

            # Check table name
            # Should not raise exceptions
            handler.handle_delete('test_table', where=None, delete_all=True)
        """
        self.handle_read(dataset)
        if not delete_all and where_list is None:
            raise HTTPException(status_code=400, detail='Parameter where is required')
        else:
            self.handle_where(where_list)

    def handle_permissions(self, dataset):
        """
        Handle a permitted table access.
        
        Parameters
        ----------
        dataset : str
            Name of the table to check.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            from msdss_base_database import Database
            from msdss_data_api.handlers import *
            from msdss_data_api.managers import *
            
            # Setup objects
            db = Database()
            dm = DataManager(database=db)
            handler = DataHandler(database=db)

            # Check if the table exists and drop if it does
            if db.has_table('test_table'):
                db.drop_table('test_table')

            # Create sample data
            data = [
                {'id': 1, 'column_one': 'a', 'column_two': 2},
                {'id': 2, 'column_one': 'b', 'column_two': 4},
                {'id': 3, 'column_one': 'c', 'column_two': 6},
            ]
            dm.create('test_table', data)

            # Check table name
            # Should not raise exceptions
            handler.handle_permissions('test_table')
        """
        if len(self.permitted_tables) > 0:
            if dataset not in self.permitted_tables:
                raise HTTPException(status_code=401)

    def handle_read(self, dataset):
        """
        Handle a table read, checking for existence and restrictions.
        
        Parameters
        ----------
        dataset : str
            Name of the table to check.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            from msdss_base_database import Database
            from msdss_data_api.handlers import *
            from msdss_data_api.managers import *
            
            # Setup objects
            db = Database()
            dm = DataManager(database=db)
            handler = DataHandler(database=db)

            # Check if the table exists and drop if it does
            if db.has_table('test_table'):
                db.drop_table('test_table')

            # Create sample data
            data = [
                {'id': 1, 'column_one': 'a', 'column_two': 2},
                {'id': 2, 'column_one': 'b', 'column_two': 4},
                {'id': 3, 'column_one': 'c', 'column_two': 6},
            ]
            dm.create('test_table', data)

            # Check table name
            # Should not raise exceptions
            handler.handle_read('test_table')
        """
        self.handle_restrictions(dataset)
        self.handle_permissions(dataset)
        if not self.database.has_table(dataset):
            raise HTTPException(status_code=404, detail='Dataset not found')

    def handle_restrictions(self, dataset):
        """
        Handle restricted table access.
        
        Parameters
        ----------
        dataset : str
            Name of the table to check.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            from msdss_base_database import Database
            from msdss_data_api.handlers import *
            from msdss_data_api.managers import *
            
            # Setup objects
            db = Database()
            dm = DataManager(database=db)
            handler = DataHandler(database=db)

            # Check if the table exists and drop if it does
            if db.has_table('test_table'):
                db.drop_table('test_table')

            # Create sample data
            data = [
                {'id': 1, 'column_one': 'a', 'column_two': 2},
                {'id': 2, 'column_one': 'b', 'column_two': 4},
                {'id': 3, 'column_one': 'c', 'column_two': 6},
            ]
            dm.create('test_table', data)

            # Check table name
            # Should not raise exceptions
            handler.handle_restrictions('test_table')
        """
        if dataset in self.restricted_tables:
            raise HTTPException(status_code=401)

    def handle_where(self, where_list):
        """
        Throw an exception if where statements do not match the expected format.
        
        Parameters
        ----------
        table : str
            Table name to apply where to.
        where_list : list(list(str))
            list of where statements the form of ``['column', 'operator', 'value']`` to further filter individual values or rows.
            
            * Operators are one of:

                .. jupyter-execute::
                    :hide-code:

                    from msdss_base_database.defaults import DEFAULT_SUPPORTED_OPERATORS
                    for operator in DEFAULT_SUPPORTED_OPERATORS:
                        print(operator)

            * Example: ``['column_two', '<'. '3']'``
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            from msdss_data_api.handlers import *

            handler = DataHandler(database=db)
            where_list = [['col', '<', '3'], ['col', '=', 'a']]

            # Should not raise exceptions
            handler.handle_where(where_list)
        """
        if where_list:

            # (DataHandler_handle_where_len) Handle wrong length
            where_has_wrong_len = any([len(w) != 3 for w in where_list])
            if where_has_wrong_len:
                raise HTTPException(status_code=400, detail='Parameter where is formatted incorrectly - should be in the form of "column operator value" e.g. "col < 3"')

            # (DataHandler_handle_where_op) Handle unsupported operator
            for w in where_list:
                if w[1].upper() not in DEFAULT_SUPPORTED_OPERATORS:
                    raise HTTPException(status_code=400, detail='Operator \'' + w[1] + '\' is not supported - supported operators are: ' + ', '.join(DEFAULT_SUPPORTED_OPERATORS))

    def handle_write(self, dataset):
        """
        Handle a table write, checking for existence and restrictions.
        
        Parameters
        ----------
        dataset : str
            Name of the table to check.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            from msdss_base_database import Database
            from msdss_data_api.handlers import *
            from msdss_data_api.managers import *
            
            # Setup objects
            db = Database()
            dm = DataManager(database=db)
            handler = DataHandler(database=db)

            # Check if the table exists and drop if it does
            if db.has_table('test_table'):
                db.drop_table('test_table')

            # Check table name
            # Should not raise exceptions
            handler.handle_write('test_table')
        """
        self.handle_restrictions(dataset)
        self.handle_permissions(dataset)
        if self.database.has_table(dataset):
            raise HTTPException(status_code=400, detail='Dataset already exists')

    def handle_update(self, dataset, data, where_list=None):
        """
        Handle a table update, checking for existence and restrictions.
        
        Parameters
        ----------
        dataset : str
            Name of the table to check.
        data : dict
            Dictionary of update values to check. Each key should represent a column name in the table.
        where_list : list(list(str))
            list of where statements the form of ``['column', 'operator', 'value']`` to further filter individual values or rows.
            
            * Operators are one of:

                .. jupyter-execute::
                    :hide-code:

                    from msdss_base_database.defaults import DEFAULT_SUPPORTED_OPERATORS
                    for operator in DEFAULT_SUPPORTED_OPERATORS:
                        print(operator)

            * Example: ``['column_two', '<'. '3']'``
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            from msdss_base_database import Database
            from msdss_data_api.handlers import *
            from msdss_data_api.managers import *
            
            # Setup objects
            db = Database()
            dm = DataManager(database=db)
            handler = DataHandler(database=db)

            # Check if the table exists and drop if it does
            if db.has_table('test_table'):
                db.drop_table('test_table')

            # Create sample data
            data = [
                {'id': 1, 'column_one': 'a', 'column_two': 2},
                {'id': 2, 'column_one': 'b', 'column_two': 4},
                {'id': 3, 'column_one': 'c', 'column_two': 6},
            ]
            dm.create('test_table', data)

            # Handle update
            # Should not throw any exceptions
            newdata = {'id': 2, 'column_one': 'UPDATED'}
            handler.handle_update('test_table', newdata)
        """
        self.handle_restrictions(dataset)
        self.handle_permissions(dataset)
        self.handle_read(dataset)
        self.handle_where(where_list)
        columns = self.database._get_table(dataset).c
        for k in data.keys():
            if k not in columns:
                raise HTTPException(status_code=400, detail='column \'' + k + '\' not found')


import argparse
import ast
import os
import pandas

from datetime import datetime
from msdss_base_database import Database
from msdss_base_database.env import DatabaseDotEnv
from msdss_base_database.defaults import DEFAULT_SUPPORTED_OPERATORS

from .core import DataAPI
from .defaults import *
from .managers import *

try:
    from msdss_users_api import UsersAPI
except ImportError:
    pass

def _get_parser():
    """
    Builds an ``argparse`` parser for the ``msdss-data`` command line tool.

    Returns
    -------
    :class:`argparse.ArgumentParser`
        An ``argparse`` parser for ``msdss-data``.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------

    .. jupyter-execute::
        :hide-output:

        from msdss_data_api.cli import _get_parser

        parser = _get_parser()
        parser.print_help()
    """

    # (_get_parser_parsers) Create main parser and sub parsers
    parser = argparse.ArgumentParser(description='Manages data with a database')
    subparsers = parser.add_subparsers(title='commands', dest='command')

    # (_get_parser_create) Add create command
    create_parser = subparsers.add_parser('create', help='create a dataset')
    create_parser.add_argument('dataset', type=str, help='dataset name')
    create_parser.add_argument('data', type=str, help='file path for data to create - supports .csv, .xlsx, .json, .xml')
    create_parser.add_argument('--metadata', metavar=('COL', 'VALUE'), nargs=2, action='append', help='metadata columns to set - e.g. title "Some Title", description "Some descr.."')

    # (_get_parser_delete) Add delete command
    delete_parser = subparsers.add_parser('delete', help='delete a dataset')
    delete_parser.add_argument('dataset', type=str, help='dataset name')
    delete_parser.add_argument('--where', action='append', help='where statement for filtering rows to delete in the form of "col operator value" (e.g. "col_a = 1") - operators supported are' + ', '.join(DEFAULT_SUPPORTED_OPERATORS))
    delete_parser.add_argument('--where_boolean', type=str, default='AND', help='bool to join multiple where statements')
    delete_parser.add_argument('--delete_all', dest='delete_all', action='store_true', help='flag to delete the entire dataset')
    delete_parser.set_defaults(delete_all=False)

    # (_get_parser_get) Add get command
    get_parser = subparsers.add_parser('get', help='query a dataset')
    get_parser.add_argument('dataset', type=str, help='dataset name')
    get_parser.add_argument('--select', action='append', help='column names to select')
    get_parser.add_argument('--where', action='append', help='where statement for filtering rows to get in the form of "col operator value" (e.g. "col_a = 1") - operators supported are' + ', '.join(DEFAULT_SUPPORTED_OPERATORS))
    get_parser.add_argument('--group_by', action='append', help='column names to group by')
    get_parser.add_argument('--aggregate', action='append', help='column names to aggregate')
    get_parser.add_argument('--aggregate_func', action='append', help='aggregate func name (e.g. sum, count) in the order of --aggregate')
    get_parser.add_argument('--order_by', action='append', help='column names to sort by')
    get_parser.add_argument('--order_by_sort', action='append', help='asc or desc for each col in --order_by')
    get_parser.add_argument('--limit', type=int, default=None, help='')
    get_parser.add_argument('--where_boolean', type=str, default='AND', help='bool to join multiple where statements')
    
    # (_get_parser_insert) Add insert command
    insert_parser = subparsers.add_parser('insert', help='insert data into a dataset')
    insert_parser.add_argument('dataset', type=str, help='dataset name')
    insert_parser.add_argument('data', type=str, help='data to insert - e.g. [ {"col_a": 1, "col_b": "a"}, {"col_a": 2, "col_b": "b"} ]')
    
    # (_get_parser_update) Add update command
    update_parser = subparsers.add_parser('update', help='update data in a dataset')
    update_parser.add_argument('--data', required=True, metavar=('COL', 'VALUE'), nargs=2, action='append', help='column names and values to update with')
    update_parser.add_argument('--where', action='append', help='where statement for filtering rows to update in the form of "col operator value" (e.g. "col_a = 1") - operators supported are' + ', '.join(DEFAULT_SUPPORTED_OPERATORS))

    # (_get_parser_get) Add get command
    search_parser = subparsers.add_parser('search', help='search datasets')
    search_parser.add_argument('--select', action='append', help='column names to select')
    search_parser.add_argument('--where', action='append', help='where statement for filtering rows to get in the form of "col operator value" (e.g. "col_a = 1") - operators supported are' + ', '.join(DEFAULT_SUPPORTED_OPERATORS))
    search_parser.add_argument('--group_by', action='append', help='column names to group by')
    search_parser.add_argument('--aggregate', action='append', help='column names to aggregate')
    search_parser.add_argument('--aggregate_func', action='append', help='aggregate func name (e.g. sum, count) in the order of --aggregate')
    search_parser.add_argument('--order_by', action='append', help='column names to sort by')
    search_parser.add_argument('--order_by_sort', action='append', help='asc or desc for each col in --order_by')
    search_parser.add_argument('--limit', type=int, default=None, help='')
    search_parser.add_argument('--where_boolean', type=str, default='AND', help='bool to join multiple where statements')

    # (_get_parser_start) Add start command
    start_parser = subparsers.add_parser('start', help='start a data api server')
    start_parser.add_argument('--host', type=str, default='127.0.0.1', help='address to host server')
    start_parser.add_argument('--port', type=int, default=8000, help='port to host server')
    start_parser.add_argument('--log_level', type=str, default='info', help='level of verbose messages to display')
    start_parser.add_argument('--set', metavar=('ROUTE', 'KEY', 'VALUE'), nargs=3, action='append', help='set route settings, where ROUTE is the route name (create, delete, metadata, etc), KEY is the setting name (e.g. path, _enable, etc), and VALUE is value for the setting')
    start_parser.add_argument('--enable_users', dest='enable_users', action='store_true', help='enable user auth for routes')
    start_parser.set_defaults(enable_users=False)

    # (_get_parser_file_key) Add file and key arguments to all commands
    for p in [parser, create_parser, delete_parser, get_parser, insert_parser, update_parser, search_parser, start_parser]:
        p.add_argument('--env_file', type=str, default='./.env', help='path of .env file')
        p.add_argument('--key_path', type=str, default=None, help='path of key file')
    
    # (_get_parser_out) Return the parser
    out = parser
    return out

def _parse_route_settings(cli_route_settings):
    """
    Parses ``--set`` route settings for the ``msdss-data`` command line tool.

    Parameters
    ----------
    cli_route_settings : list(tuple)
        List of tuples of length 3, representing the route, key setting and value setting in order.

    Returns
    -------
    dict
        A dictionary that can be passed to parameter ``route_settings`` in :func:`msdss_data_api.routers.get_data_router`

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------

    .. jupyter-execute::
        :hide-output:

        from msdss_data_api.cli import _parse_route_settings
        from pprint import pprint

        cli_route_settings = [
            ('create', '_restricted_tables', '["user", "data"]'),
            ('create', '_enable', 'True'),
            ('create', '_get_user', '{"superuser": True}'),
            ('delete', 'path', '/{dataset}'),
            ('delete', 'tags', '["data"]')
        ]
        route_settings = _parse_route_settings(cli_route_settings)
        pprint(route_settings)
    """
    out = {route: {} for route, k, v in cli_route_settings}
    for route, key, value in cli_route_settings:
        if key in ('tags', '_restricted_tables', '_get_user', '_enable'):
            out[route][key] = ast.literal_eval(value)
        else:
            out[route][key] = value
    return out

def _read_data_file(data_path):
    """
    Reads a data file into a :class:`pandas:pandas.DataFrame` object.

    Parameters
    ----------
    data_path : str
        Path of the data file with extension. Supports ``.csv``, ``.xlsx``/``.xls``, ``.json``, and ``.xml``.

    Returns
    -------
    :class:`pandas:pandas.DataFrame`
        A dataframe object of the file data.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------

    .. code::

        from msdss_data_api.cli import _read_data_file

        data = _read_data_file('path/to/data.json')
    """
    data_ext = os.path.splitext(data_path)[1].lower()
    if data_ext == '.csv':
        out = pandas.read_csv(data_path)
    elif data_ext in ('.xlsx', '.xls'):
        out = pandas.read_excel(data_path)
    elif data_ext == '.json':
        out = pandas.read_json(data_path)
    elif data_ext == '.xml':
        out = pandas.read_xml(data_path)
    return out

def run():
    """
    Runs the ``msdss-data`` command.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------
    >>> msdss-data --help

    .. jupyter-execute::
        :hide-code:

        from msdss_data_api.cli import _get_parser

        parser = _get_parser()
        parser.print_help()

    Create a dataset:

    >>> msdss-data create <data_name> <path/to/data.json> --metadata title "Test Data"

    Query dataset:

    >>> msdss-data get <data_name>

    Update dataset values:

    >>> msdss-data update <data_name> --data <col_name> <update_value> --where <col_name> = <value>

    Insert data into dataset:

    >>> msdss-data insert <data_name> "[{'col_a': 1, 'col_b': 'a'}]"

    Delete a dataset:

    >>> msdss-data delete <data_name> --delete_all

    Search datasets:

    >>> msdss-data search --where "dataset = <data_name>"

    Start an API server:

    >>> msdss-data start
    """

    # (run_kwargs) Get arguments and command
    parser = _get_parser()
    kwargs = vars(parser.parse_args())
    command = kwargs.pop('command')

    # (run_env) Get env vars
    env_kwargs = dict(
        env_file=kwargs.pop('env_file'),
        key_path=kwargs.pop('key_path')
    )

    # (run_database) Create database obj
    database = Database(env=DatabaseDotEnv(**env_kwargs))

    # (run_managers) Create data and metadata managers
    data_manager = DataManager(database=database)
    metadata_manager = MetadataManager(data_manager=data_manager)

    # (run_command) Run commands
    if command == 'create':

        # (run_command_create_metadata) Get metadata attrs 
        metadata = kwargs.pop('metadata', None)
        metadata = dict(metadata) if metadata else {}
        metadata['created_by'] = 'admin'
        metadata['created_at'] = datetime.now()
        metadata['updated_at'] = datetime.now()

        # (run_command_create_read) Read data from file
        kwargs['data'] = _read_data_file(kwargs.pop('data'))

        # (run_command_create_run) Create the dataset
        try:
            dataset = kwargs.get('dataset')
            data_manager.create(**kwargs)
            metadata_manager.create(dataset, metadata)
            print('Dataset created ' + dataset)
        except HTTPException as e:
            print(e.detail)

    elif command == 'delete':

        # (run_command_delete) Execute user delete
        try:
            dataset = kwargs.get('dataset')
            data_manager.delete(**kwargs)
            metadata_manager.delete(dataset)
            print('Dataset deleted ' + dataset)
        except HTTPException as e:
            print(e.detail)

    elif command == 'get':

        # (run_command_get) Query a dataset
        try:
            dataset = kwargs.get('dataset')
            data = data_manager.get(**kwargs)
            print(pandas.DataFrame(data))
        except HTTPException as e:
            print(e.detail)

    elif command == 'insert':

        # (run_command_insert) Insert data into a dataset
        try:
            kwargs['data'] = ast.literal_eval(kwargs['data'])
            dataset = kwargs.get('dataset')
            data_manager.insert(**kwargs)
            metadata_manager.updated_at(dataset)
            print('Dataset insert ' + dataset)
        except HTTPException as e:
            print(e.detail)

    elif command == 'update':

        # (run_command_update_data) Get data attrs 
        kwargs['data'] = dict(kwargs['data'])

        # (run_command_update_run) Update dataset data
        try:
            dataset = kwargs.get('dataset')
            data_manager.update(**kwargs)
            metadata_manager.updated_at(dataset)
            print('Dataset update ' + dataset)
        except HTTPException as e:
            print(e.detail)

    elif command == 'search':

        # (run_command_search_run) Search datasets
        try:
            data = metadata_manager.search(**kwargs)
            print(pandas.DataFrame(data))
        except HTTPException as e:
            print(e.detail)

    elif command == 'start':

        # (run_command_start_server) Extract server kwargs
        start_kwargs = dict(
            host=kwargs.pop('host'),
            port=kwargs.pop('port'),
            log_level=kwargs.pop('log_level')
        )

        # (run_command_start_env) Configure env database and users api
        kwargs['database'] = database
        enable_users = kwargs.pop('enable_users')
        if enable_users:
            from msdss_users_api import UsersAPI
            from msdss_users_api.env import UsersDotEnv
            kwargs['users_api'] = UsersAPI(env=UsersDotEnv(**env_kwargs))
        else:
            kwargs['users_api'] = None

        # (run_command_start_settings) Convert route settings
        cli_route_settings = kwargs.pop('set', None)
        kwargs['data_router_settings'] = dict(
            route_settings=_parse_route_settings(cli_route_settings) if cli_route_settings else {}
        )

        # (run_command_start) Execute users api server
        app = DataAPI(**kwargs)
        app.start(**start_kwargs)


DEFAULT_METADATA_TABLE = 'data'
DEFAULT_USER_TABLE = 'user'
DEFAULT_NAME_COLUMN = 'dataset'
DEFAULT_UPDATE_COLUMN = 'updated_at'
DEFAULT_RESTRICTED_TABLES = [DEFAULT_METADATA_TABLE, DEFAULT_USER_TABLE]
DEFAULT_METADATA_COLUMNS = [
    dict(name='id', type_='Integer', primary_key=True),
    dict(name='dataset', type_='String', unique=True),
    ('title', 'String'),
    ('description', 'String'),
    ('tags', 'String'),
    ('source', 'String'),
    ('created_by', 'String'),
    ('created_at', 'DateTime'),
    ('updated_at', 'DateTime')
]
DEFAULT_DATA_ROUTE_SETTINGS = dict(
    columns=dict(
        path='/{dataset}/columns',
        _restricted_tables=DEFAULT_RESTRICTED_TABLES,
        _enable=True,
        _get_user={}
    ),
    create=dict(
        path='/',
        _restricted_tables=DEFAULT_RESTRICTED_TABLES,
        _enable=True,
        _get_user={'superuser': True}
    ),
    delete=dict(
        path='/{dataset}',
        _restricted_tables=DEFAULT_RESTRICTED_TABLES,
        _enable=True,
        _get_user={'superuser': True}
    ),
    id=dict(
        path='/{dataset}/id/{id}',
        _restricted_tables=DEFAULT_RESTRICTED_TABLES,
        _enable=True,
        _get_user={}
    ),
    insert=dict(
        path='/{dataset}/insert',
        _restricted_tables=DEFAULT_RESTRICTED_TABLES,
        _enable=True,
        _get_user={'superuser': True}
    ),
    metadata=dict(
        path='/{dataset}/metadata',
        tags=['metadata'],
        _restricted_tables=DEFAULT_RESTRICTED_TABLES,
        _enable=True,
        _get_user={}
    ),
    metadata_update=dict(
        path='/{dataset}/metadata',
        tags=['metadata'],
        _restricted_tables=DEFAULT_RESTRICTED_TABLES,
        _enable=True,
        _get_user={'superuser': True}
    ),
    query=dict(
        path='/{dataset}',
        _restricted_tables=DEFAULT_RESTRICTED_TABLES,
        _enable=True,
        _get_user={}
    ),
    rows=dict(
        path='/{dataset}/rows',
        _restricted_tables=DEFAULT_RESTRICTED_TABLES,
        _enable=True,
        _get_user={}
    ),
    search=dict(
        path='/',
        _restricted_tables=DEFAULT_RESTRICTED_TABLES,
        _enable=True,
        _get_user={}
    ),
    update=dict(
        path='/{dataset}',
        _restricted_tables=DEFAULT_RESTRICTED_TABLES,
        _enable=True,
        _get_user={'superuser': True}
    )
)


from msdss_base_database import Database

from .handlers import *
from .managers import *

def create_data_manager_func(database=None, permitted_tables=[], restricted_tables=DEFAULT_RESTRICTED_TABLES):
    """
    Create a function yielding a :class:`msdss_data_api.managers.DataManager`.
    
    Parameters
    ----------
    database : :class:`msdss_base_database:msdss_base_database.core.Database` or None
        Database object to use for APIs. If ``None``, a default database will be used.
    permitted_tables : list(str)
        List of permitted tables in the database that can only be accessed.
    restricted_tables : list(str)
        List of restricted tables in the database to prevent access to.

    Returns
    -------
    func
        A function yielding a preconfigured :class:`msdss_data_api.managers.DataManager`.
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from msdss_base_database import Database
        from msdss_data_api.tools import *
        
        # Setup database
        db = Database()

        # Create a function yielding the data manager to use as a dependency
        get_data_manager = create_data_manager_func(database=db)
    """
    database = database if database else Database()
    data_handler = DataHandler(database=database, permitted_tables=permitted_tables, restricted_tables=restricted_tables)
    data_manager = DataManager(database=database, handler=data_handler)
    async def out():
        yield data_manager
    return out

def create_metadata_manager_func(database=None):
    """
    Create a function yielding a :class:`msdss_data_api.managers.MetadataManager`.
    
    Parameters
    ----------
    database : :class:`msdss_base_database:msdss_base_database.core.Database` or None
        Database object to use for APIs. If ``None``, a default database will be used.

    Returns
    -------
    func
        A function yielding a preconfigured :class:`msdss_data_api.managers.MetadataManager`.
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from msdss_base_database import Database
        from msdss_data_api.tools import *
        
        # Setup database
        db = Database()

        # Create a function yielding the metadata manager to use as a dependency
        get_metadata_manager = create_metadata_manager_func()
    """
    database = database if database else Database()
    data_manager = DataManager(database=database)
    metadata_manager = MetadataManager(data_manager=data_manager)
    async def out():
        yield metadata_manager
    return out


from fastapi import FastAPI
from msdss_base_api import API
from msdss_base_database import Database

from .routers import *
from .handlers import *

class DataAPI(API):
    """
    Class for creating Data APIs.
    
    Parameters
    ----------
    users_api : :class:`msdss_users_api:msdss_users_api.core.UsersAPI` or None
        Users API object to enable user authentication for data routes.
        If ``None``, user authentication will not be used for data routes.
    database : :class:`msdss_base_database:msdss_base_database.core.Database` or None
        A :class:`msdss_base_database:msdss_base_database.core.Database` object for managing data. If ``None``, a default database will be setup.
    data_router_settings : dict
        Additional arguments passed to :func:`msdss_data_api.routers.get_data_router` except ``database``.
    api : :class:`fastapi:fastapi.FastAPI`
        API object for creating routes.
    *args, **kwargs
        Additional arguments passed to :class:`msdss_base_api:msdss_base_api.core.API`.

    Attributes
    ----------
    data_api_database : :class:`msdss_base_datbase:msdss_base_database.core.Database`
        Database object used for the data API.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    Create Data API without users:

    .. jupyter-execute::

        from msdss_base_database import Database
        from msdss_data_api import DataAPI
        from msdss_users_api import UsersAPI

        # Create database object
        database = Database(
            driver='postgresql',
            user='msdss',
            password='msdss123',
            host='localhost',
            port='5432',
            database='msdss'
        )

        # Create a data api without users
        app = DataAPI(database=database)

    Create Data API with users:

    .. jupyter-execute::

        from msdss_base_database import Database
        from msdss_data_api import DataAPI
        from msdss_users_api import UsersAPI

        # Create database object
        database = Database(
            driver='postgresql',
            user='msdss',
            password='msdss123',
            host='localhost',
            port='5432',
            database='msdss'
        )

        # Create a data api with users
        # CHANGE SECRETS TO STRONG PHRASES
        users_api = UsersAPI(
            'cookie-secret',
            'jwt-secret',
            'reset-secret',
            'verification-secret',
            database=database
        )
        app = DataAPI(users_api, database=database)

        # Add users routes
        app.add_apps(users_api)

        # Run the app with app.start()
        # Try API at http://localhost:8000/docs
        # app.start()
    """
    def __init__(
        self,
        users_api=None,
        database=None,
        data_router_settings={},
        api=FastAPI(
            title='MSDSS Data API',
            version='0.2.8'
        ),
        *args, **kwargs):
        super().__init__(api=api, *args, **kwargs)

        # (DataAPI_settings) Setup router params
        database = database if database else Database()
        data_router_settings['database'] = database
        
        # (DataAPI_users) Add users app if specified
        if users_api:
            data_router_settings['users_api'] = users_api
        
        # (DataAPI_router_data) Add data router
        data_router = get_data_router(**data_router_settings)
        self.add_router(data_router)

        # (DataAPI_attr) Set attributes
        self.data_api_database = database
        


from datetime import datetime
from fastapi import HTTPException
from msdss_base_database import Database
from shlex import split

from .defaults import *
from .handlers import *

class DataManager:
    """
    Class to manage datasets in a database.
    
    Parameters
    ----------
    database : :class:`msdss_base_database:msdss_base_database.core.Database` or None
        Database object to use for creating data. If ``None``, a default database will be used.
    handler : :class:`msdss_data_api.data.DataHandler` or bool or None
        Handler for handling dataset events.
        
        * If ``None``, a default handler will be created with parameter ``database``
        * If ``False``, events will not be handled
        * Sets the handler database to be the parameter ``database``
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from msdss_base_database import Database
        from msdss_data_api.managers import *
        
        # Setup database
        db = Database()
        dm = DataManager(database=db)

        # Check if the table exists and drop if it does
        if db.has_table("test_table"):
            db.drop_table("test_table")

        # Create sample data
        data = {
            'id': [1, 2, 3],
            'column_one': ['a', 'b', 'c'],
            'column_two': [2, 4, 6]
        }
        dm.create('test_table', data)

        # Query sample data
        result = dm.get('test_table')

        # Update sample data
        new_data = {'column_one': 'updated_at'}
        dm.update('test_table', new_data, where=['id > 1'])

        # Delete sample data
        dm.delete('test_table', where=['id = 1'])

        # Delete the entire dataset
        dm.delete('test_table', delete_all=True)
    """
    def __init__(self, database=None, handler=None):
        database = database if database else Database()
        handler = DataHandler(database=database) if handler is None else handler
        self.database = database
        self.handler = handler
        if self.handler:
            self.handler.database = database

    def create(self, name, data):
        """
        Create a dataset.

        See :meth:`msdss_base_database:msdss_base_database.core.Database.insert`.
        
        Parameters
        ----------
        name : str
            Name of the dataset or table to hold the data.
        data : list(dict)
            Data to insert into the table. Should be a list of dictionaries with the same keys, where each key in each dict is a column name.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            from msdss_base_database import Database
            from msdss_data_api.managers import *
            
            # Setup database
            db = Database()
            dm = DataManager(database=db)

            # Check if the table exists and drop if it does
            if db.has_table("test_table"):
                db.drop_table("test_table")

            # Create sample data
            data = [
                {'id': 1, 'column_one': 'a', 'column_two': 2},
                {'id': 2, 'column_one': 'b', 'column_two': 4},
                {'id': 3, 'column_one': 'c', 'column_two': 6},
            ]
            dm.create('test_table', data)
        """
        if self.handler:
            self.handler.handle_restrictions(name)
            self.handler.handle_write(name)
        self.database.insert(name, data)

    def delete(self, name, where=None, where_boolean='AND', delete_all=False):
        """
        Delete a dataset.

        See :meth:`msdss_base_database:msdss_base_database.core.Database.delete`.
        
        Parameters
        ----------
        name : str
            Name of the dataset or table in the database to delete.
        where : list(str)
            list of where statements the form of ``column operator value`` to further filter individual values or rows for deleting.
            
            * Operators are one of

                .. jupyter-execute::
                    :hide-code:

                    from msdss_base_database.defaults import DEFAULT_SUPPORTED_OPERATORS
                    for operator in DEFAULT_SUPPORTED_OPERATORS:
                        print(operator)

            * Example: ``'column_two < 3'``

        where_boolean : str
            One of AND or OR to combine where statements with. Defaults to AND if not one of AND or OR.
        delete_all : bool
            Whether to remove the entire dataset or not.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            from msdss_base_database import Database
            from msdss_data_api.managers import *
            
            # Setup database
            db = Database()
            dm = DataManager(database=db)

            # Check if the table exists and drop if it does
            if db.has_table("test_table"):
                db.drop_table("test_table")

            # Create sample data
            data = [
                {'id': 1, 'column_one': 'a', 'column_two': 2},
                {'id': 2, 'column_one': 'b', 'column_two': 4},
                {'id': 3, 'column_one': 'c', 'column_two': 6},
            ]
            dm.create('test_table', data)

            # Delete sample data
            dm.delete('test_table', where=['id = 1'])
            res = dm.get('test_table')
            print(res)

            # Delete the entire dataset
            dm.delete('test_table', delete_all=True)
        """

        # (DataManager_delete_handle) Handle delete operation on data
        where = [split(w) for w in where] if where else where
        if self.handler:
            self.handler.handle_delete(name, where, delete_all)

        # (DataManager_delete_run) Delete data
        if delete_all:
            self.database.drop_table(name)
        else:
            self.database.delete(name, where=where, where_boolean=where_boolean)

    def get(
        self,
        name,
        select='*',
        where=None,
        group_by=None,
        aggregate=None,
        aggregate_func=None,
        order_by=None,
        order_by_sort=None,
        limit=None,
        offset=None,
        where_boolean='AND',
        *args, **kwargs):
        """
        Query data from the database.

        See :meth:`msdss_base_database:msdss_base_database.core.Database.select`.
        
        Parameters
        ----------
        name : str
            Name of the database table to query from.
        select : list(str) or None
            List of column names or a single column name to filter or select from the table.
            
            * If ``None``, columns will not be added to the select statement.
            * If ``'*'``, then all columns will be selected
            
        where : list(str)
            list of where statements the form of ``column operator value`` to further filter individual values or rows.
            
            * Operators are one of

                .. jupyter-execute::
                    :hide-code:

                    from msdss_base_database.defaults import DEFAULT_SUPPORTED_OPERATORS
                    for operator in DEFAULT_SUPPORTED_OPERATORS:
                        print(operator)

            * Example: ``'column_two < 3'``
        
        group_by : list(str) or None
            Single or list of column names to group by. This should be used with ``aggregate`` and ``aggregate_func``.
        aggregate : list(str) or None
            Single or list of column names to aggregate using the ``aggregate_func``. This should be used with ``group_by`` and ``aggregate_func``.
        aggregate_func : list(str)
            Function name (such as 'count' or 'sum') from :class:`sqlalchemy:sqlalchemy.sql.functions.Function` for aggregating records from each ``aggregate`` column.
            If a list of str, then it must have the same number of elements as ``aggregate`` or else only the shortest length list will be used.
        order_by : list(str) or None
            Single or list of column names to order or sort by.
        order_by_sort : list(str)
            Sort the records in increasing or decreasing order by each column in ``order_by``, where the value can be one of 'asc' for ascending or 'desc' for descending'.
            If a list of str, then it must have the same number of elements as ``order_by`` or else only the shortest length list will be used.
        limit : int or None
            Integer number to limit the number of rows returned.
        offset : int or None
            Number of rows to skip.
        where_boolean : str
            One of ``AND`` or ``OR`` to combine ``where`` statements with. Defaults to ``AND`` if not one of ``AND`` or ``OR``.
        *args, **kwargs
            Additional arguments passed to :meth:`msdss_base_database:msdss_base_database.core.Database.select`.
        
        Returns
        -------
        list(dict)
            A list of dicts where each key is the column name.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>

        Example
        -------
        .. jupyter-execute::

            from msdss_base_database import Database
            from msdss_data_api.managers import *
            
            # Setup objects
            db = Database()
            dm = DataManager(database=db)

            # Check if the table exists and drop if it does
            if db.has_table("test_table"):
                db.drop_table("test_table")

            # Create sample data
            data = [
                {'id': 1, 'column_one': 'a', 'column_two': 2},
                {'id': 2, 'column_one': 'b', 'column_two': 4},
                {'id': 3, 'column_one': 'c', 'column_two': 6},
            ]
            dm.create('test_table', data)

            # Query the data from the database
            result = dm.get('test_table')
            print(result)
        """

        # (DataManager_get_handle) Handle get operation on data
        where = [split(w) for w in where] if where else where
        if self.handler:
            self.handler.handle_read(name)
            self.handler.handle_where(where)

        # (DataManager_get_run) Query the database
        out = self.database.select(
            table=name,
            select=select,
            where=where,
            group_by=group_by,
            aggregate=aggregate,
            aggregate_func=aggregate_func,
            order_by=order_by,
            order_by_sort=order_by_sort,
            limit=limit,
            offset=offset,
            where_boolean=where_boolean,
            *args, **kwargs
        ).to_dict(orient='records')
        return out

    def get_columns(self, name):
        """
        Get number of columns for a dataset.

        See :meth:`msdss_base_database:msdss_base_database.core.Database.columns`.
        
        Parameters
        ----------
        name : str
            Name of the dataset or table in the database.

        Returns
        -------
        int
            Number of columns for the dataset.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            from msdss_base_database import Database
            from msdss_data_api.managers import *
            
            # Setup database
            db = Database()
            dm = DataManager(database=db)

            # Check if the table exists and drop if it does
            if db.has_table("test_table"):
                db.drop_table("test_table")

            # Create sample data
            data = [
                {'id': 1, 'column_one': 'a', 'column_two': 2},
                {'id': 2, 'column_one': 'b', 'column_two': 4},
                {'id': 3, 'column_one': 'c', 'column_two': 6},
            ]
            dm.create('test_table', data)

            # Get num of cols
            columns = dm.get_columns('test_table')
            print(f'columns: {columns}')
        """
        if self.handler:
            self.handler.handle_read(name)
        out = self.database.columns(name)
        return out

    def get_rows(self, name):
        """
        Get number of rows for a dataset.

        See :meth:`msdss_base_database:msdss_base_database.core.Database.rows`.
        
        Parameters
        ----------
        name : str
            Name of the dataset or table in the database.

        Returns
        -------
        int
            Number of rows for the dataset.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            from msdss_base_database import Database
            from msdss_data_api.managers import *
            
            # Setup database
            db = Database()
            dm = DataManager(database=db)

            # Check if the table exists and drop if it does
            if db.has_table("test_table"):
                db.drop_table("test_table")

            # Create sample data
            data = [
                {'id': 1, 'column_one': 'a', 'column_two': 2},
                {'id': 2, 'column_one': 'b', 'column_two': 4},
                {'id': 3, 'column_one': 'c', 'column_two': 6},
            ]
            dm.create('test_table', data)

            # Get num of rows
            rows = dm.get_rows('test_table')
            print(f'rows: {rows}')
        """
        if self.handler:
            self.handler.handle_read(name)
        out = self.database.rows(name)
        return out

    def insert(self, name, data):
        """
        Create a dataset.

        See :meth:`msdss_base_database:msdss_base_database.core.Database.insert`.
        
        Parameters
        ----------
        name : str
            Name of the dataset or table to hold the data.
        data : list(dict)
            Data to insert into the table. Should be a list of dictionaries with the same keys, where each key in each dict is a column name.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            from msdss_base_database import Database
            from msdss_data_api.managers import *
            
            # Setup database
            db = Database()
            dm = DataManager(database=db)

            # Check if the table exists and drop if it does
            if db.has_table("test_table"):
                db.drop_table("test_table")

            # Create sample data
            data = [
                {'id': 1, 'column_one': 'a', 'column_two': 2},
                {'id': 2, 'column_one': 'b', 'column_two': 4},
                {'id': 3, 'column_one': 'c', 'column_two': 6},
            ]
            dm.create('test_table', data)

            # Insert more data
            more_data = [
                {'id': 4, 'column_one': 'e', 'column_two': 8},
                {'id': 5, 'column_one': 'f', 'column_two': 10},
                {'id': 6, 'column_one': 'g', 'column_two': 12},
            ]
            dm.insert('test_table', more_data)
        """
        if self.handler:
            self.handler.handle_restrictions(name)
            self.handler.handle_read(name)
        self.database.insert(name, data)

    def update(self, name, data, where):
        """
        Update data from the database.

        See :meth:`msdss_base_database:msdss_base_database.core.Database.update`.
        
        Parameters
        ----------
        name : str
            Name of the table to update.
        data : dict
            Dictionary representing values to update if they match the ``where`` parameter requirements. Each key is a column and the value is the updated new value.
        where : list(str)
            list of where statements the form of ``column operator value`` to further filter individual values or rows.
            
            * Operators are one of

                .. jupyter-execute::
                    :hide-code:

                    from msdss_base_database.defaults import DEFAULT_SUPPORTED_OPERATORS
                    for operator in DEFAULT_SUPPORTED_OPERATORS:
                        print(operator)
                            
            * Example: ``'column_two < 3'``
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>

        Example
        -------
        .. jupyter-execute::

            from msdss_base_database import Database
            from msdss_data_api.managers import *
            
            # Setup database
            db = Database()
            dm = DataManager(database=db)

            # Check if the table exists and drop if it does
            if db.has_table("test_table"):
                db.drop_table("test_table")

            # Create sample data
            data = [
                {'id': 1, 'column_one': 'a', 'column_two': 2},
                {'id': 2, 'column_one': 'b', 'column_two': 4},
                {'id': 3, 'column_one': 'c', 'column_two': 6},
            ]
            dm.create('test_table', data)

            # Update the data from the database
            new_data = {'column_one': 'updated_at'}
            dm.update('test_table', new_data, where=['id > 1'])

            # See updated data
            result = dm.get('test_table')
            print(result)
        """
        where = [split(w) for w in where]
        if self.handler:
            self.handler.handle_update(name, data, where)
        self.database.update(table=name, where=where, values=data)

class MetadataManager:
    """
    Class to manage metadata in a database.
    
    Parameters
    ----------
    data_manager : :class:`msdss_data_api.managers.DataManager`
        Data manager object for managing datasets in a database.
        The restricted tables for the handler will be set to ``[]`` while the only permitted table will be the table name of the parameter ``table``.
    table : str
        The name of the table to store the metadata.
    columns : list(dict) or list(list)
        List of dict (kwargs) or lists (positional args) that are passed to sqlalchemy.schema.Column. See parameter ``columns`` in :meth:`msdss_base_database:msdss_base_database.core.create_table`.
        This defines the table to store the metadata, where the default is:

        .. jupyter-execute::
            :hide-code:

            from msdss_data_api.defaults import *
            from pprint import pprint

            pprint(DEFAULT_METADATA_COLUMNS)
    name_column : str
        Name of the column identifying each entry.
    updated_column : str
        Name of the column for storing the last updated date/time.

    Attributes
    ----------
    table : str
        The name of the metadata table.
    data_manager : :class:`msdss_data_api.managers.DataManager`
        Same as parameter ``data_manager``.
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from datetime import datetime
        from msdss_base_database import Database
        from msdss_data_api.managers import *
        from msdss_data_api.defaults import *
        
        # Setup database
        db = Database()

        # Check if the metadata table exists and drop if it does
        if db.has_table(DEFAULT_METADATA_TABLE):
            db.drop_table(DEFAULT_METADATA_TABLE)

        # Setup metadata manager
        data_manager = DataManager(database=db)
        mdm = MetadataManager(data_manager)

        # Add metadata
        metadata = [{
            'title': 'Testing Data',
            'description': 'Data used for testing',
            'tags': 'test exp auto',
            'source': 'Automatically generated from Python',
            'created_by': 'msdss',
            'created_at': datetime.now(),
            'updated_at': datetime.now()
        }]
        mdm.create('test_data', metadata)

        # Get metadata
        metadata_get = mdm.get('test_data')
        
        # Search metadata
        search_results = mdm.search(where=['title = "Testing Data"'])

        # Update metadata
        mdm.update('test_data', {'description': 'NEW DESCRIPTION'})

        # Delete metadata
        mdm.delete('test_data')
    """
    def __init__(
        self,
        data_manager=DataManager(),
        table=DEFAULT_METADATA_TABLE,
        columns=DEFAULT_METADATA_COLUMNS,
        name_column=DEFAULT_NAME_COLUMN,
        updated_column=DEFAULT_UPDATE_COLUMN):
        
        # (MetadataManager_table) Create table if not exists
        if not data_manager.database.has_table(table):
            data_manager.database.create_table(table, columns)
        
        # (MetadataManager_attr) Set attributes
        self.table = table
        self.name_column = name_column
        self.updated_column = updated_column

        # (MetadataManager_manager) Setup data manager
        self.data_manager = data_manager
        self.data_manager.handler.permitted_tables = [self.table]
        self.data_manager.handler.restricted_tables = []

    def create(self, name, data):
        """
        Create a metadata entry.

        See :meth:`msdss_data_api.managers.DataManager.delete`.
        
        Parameters
        ----------
        name : str
            Name of the entry (e.g. dataset) to add metadata for.
        data : list(dict) or dict
            Metadata to insert into the table, where each key represents a metadata descriptor. The default key names are:

            .. jupyter-execute::
                :hide-code:

                from msdss_data_api.defaults import *
                from pprint import pprint

                pprint(DEFAULT_METADATA_COLUMNS)

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>

        Example
        -------
        .. jupyter-execute::

            from datetime import datetime
            from pprint import pprint
            from msdss_base_database import Database
            from msdss_data_api.managers import *
            from msdss_data_api.defaults import *
            
            # Setup database
            db = Database()

            # Check if the metadata table exists and drop if it does
            if db.has_table(DEFAULT_METADATA_TABLE):
                db.drop_table(DEFAULT_METADATA_TABLE)

            # Setup metadata manager
            data_manager = DataManager(database=db)
            mdm = MetadataManager(data_manager)

            # Add metadata
            metadata = [{
                'title': 'Testing Data',
                'description': 'Data used for testing',
                'tags': 'test exp auto',
                'source': 'Automatically generated from Python',
                'created_by': 'msdss',
                'created_at': datetime.now(),
                'updated_at': datetime.now()
            }]
            mdm.create('test_data', metadata)

            # Print results
            tb = mdm.search()
            pprint(tb)
        """
        data = [data] if isinstance(data, dict) else data
        data[0][self.name_column] = name
        self.data_manager.insert(self.table, data)

    def delete(self, name):
        """
        Delete a metadata entry.

        See :meth:`msdss_data_api.managers.DataManager.delete`.
        
        Parameters
        ----------
        name : str
            Name of the entry (e.g. dataset) to add metadata for.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>

        Example
        -------
        .. jupyter-execute::

            from datetime import datetime
            from pprint import pprint
            from msdss_base_database import Database
            from msdss_data_api.managers import *
            from msdss_data_api.defaults import *
            
            # Setup database
            db = Database()

            # Check if the metadata table exists and drop if it does
            if db.has_table(DEFAULT_METADATA_TABLE):
                db.drop_table(DEFAULT_METADATA_TABLE)

            # Setup metadata manager
            data_manager = DataManager(database=db)
            mdm = MetadataManager(data_manager)

            # Add metadata
            metadata = [{
                'dataset': 'test_data',
                'title': 'Testing Data',
                'description': 'Data used for testing',
                'tags': 'test exp auto',
                'source': 'Automatically generated from Python',
                'created_by': 'msdss',
                'created_at': datetime.now(),
                'updated_at': datetime.now()
            }]
            mdm.create('test_data', metadata)
            before_delete = mdm.search()

            # Delete metadata
            mdm.delete('test_data')
            after_delete = mdm.search()

            # Print results
            print('before_delete:\\n')
            pprint(before_delete)
            print('\\nafter_delete:\\n')
            pprint(after_delete)
        """
        where = [f'{self.name_column} = {name}']
        self.data_manager.delete(self.table, where=where)

    def get(self, name):
        """
        Search metadata entries.

        See :meth:`msdss_data_api.managers.DataManager.get`.
        
        Parameters
        ----------
        name : str
            Name of the entry (e.g. dataset) to add metadata for.

        Returns
        -------
        list(dict)
            A list of dicts where each key is the column name.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>

        Example
        -------
        .. jupyter-execute::

            from datetime import datetime
            from pprint import pprint
            from msdss_base_database import Database
            from msdss_data_api.managers import *
            from msdss_data_api.defaults import *
            
            # Setup database
            db = Database()

            # Check if the metadata table exists and drop if it does
            if db.has_table(DEFAULT_METADATA_TABLE):
                db.drop_table(DEFAULT_METADATA_TABLE)

            # Setup metadata manager
            data_manager = DataManager(database=db)
            mdm = MetadataManager(data_manager)

            # Add metadata
            metadata = [{
                'title': 'Testing Data',
                'description': 'Data used for testing',
                'tags': 'test exp auto',
                'source': 'Automatically generated from Python',
                'created_by': 'msdss',
                'created_at': datetime.now(),
                'updated_at': datetime.now()
            }]
            mdm.create('test_data', metadata)

            # Get metadata
            metadata_get = mdm.get('test_data')
            pprint(metadata_get)
        """
        where = [f'{self.name_column} = {name}']
        out = self.data_manager.get(self.table, where=where)
        return out

    def search(self, *args, **kwargs):
        """
        Search metadata entries.

        See :meth:`msdss_data_api.managers.DataManager.get`.
        
        Parameters
        ----------
        *args, **kwargs
            Additional arguments passed to :meth:`msdss_data_api.managers.DataManager.get` except for parameter ``table``.

        Returns
        -------
        list(dict)
            A dict of lists where each key is the column name and each list contains the values for columns in the order of the rows of the table.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>

        Example
        -------
        .. jupyter-execute::

            from datetime import datetime
            from pprint import pprint
            from msdss_base_database import Database
            from msdss_data_api.managers import *
            from msdss_data_api.defaults import *
            
            # Setup database
            db = Database()

            # Check if the metadata table exists and drop if it does
            if db.has_table(DEFAULT_METADATA_TABLE):
                db.drop_table(DEFAULT_METADATA_TABLE)

            # Setup metadata manager
            data_manager = DataManager(database=db)
            mdm = MetadataManager(data_manager)

            # Add metadata
            metadata = [{
                'title': 'Testing Data',
                'description': 'Data used for testing',
                'tags': 'test exp auto',
                'source': 'Automatically generated from Python',
                'created_by': 'msdss',
                'created_at': datetime.now(),
                'updated_at': datetime.now()
            }]
            mdm.create('test_data', metadata)

            # Search metadata
            results = mdm.search(where=['title = "Testing Data"'])
            pprint(results)
        """
        out = self.data_manager.get(self.table, *args, **kwargs)
        return out

    def update(self, name, data):
        """
        Update metadata entry.

        See :meth:`msdss_data_api.managers.DataManager.update`.
        
        Parameters
        ----------
        name : str
            Name of the entry (e.g. dataset) to add metadata for.
        data : dict
            Dictionary representing values to update.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>

        Example
        -------
        .. jupyter-execute::

            from datetime import datetime
            from pprint import pprint
            from msdss_base_database import Database
            from msdss_data_api.managers import *
            from msdss_data_api.defaults import *
            
            # Setup database
            db = Database()

            # Check if the metadata table exists and drop if it does
            if db.has_table(DEFAULT_METADATA_TABLE):
                db.drop_table(DEFAULT_METADATA_TABLE)

            # Setup metadata manager
            data_manager = DataManager(database=db)
            mdm = MetadataManager(data_manager)

            # Add metadata
            metadata = [{
                'title': 'Testing Data',
                'description': 'Data used for testing',
                'tags': 'test exp auto',
                'source': 'Automatically generated from Python',
                'created_by': 'msdss',
                'created_at': datetime.now(),
                'updated_at': datetime.now()
            }]
            mdm.create('test_data', metadata)
            before_update = mdm.get('test_data')

            # Update metadata
            mdm.update('test_data', {'description': 'NEW DESCRIPTION'})
            after_update = mdm.get('test_data')

            # Print results
            print('before_update:\\n')
            pprint(before_update)
            print('\\nafter_update:\\n')
            pprint(after_update)
        """
        where = [f'{self.name_column} = {name}']
        self.data_manager.update(self.table, data, where=where)

    def updated_at(self, name, dt=None):
        """
        Set last updated entry.

        See :meth:`msdss_data_api.managers.MetadataManager.update`.
        
        Parameters
        ----------
        name : str
            Name of the entry (e.g. dataset) to add metadata for.
        dt : :class:`datetime.datetime` or None
            Datetime object representing the last updated time. If ``None``, will be set to now.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>

        Example
        -------
        .. jupyter-execute::

            from datetime import datetime
            from pprint import pprint
            from msdss_base_database import Database
            from msdss_data_api.managers import *
            from msdss_data_api.defaults import *
            
            # Setup database
            db = Database()

            # Check if the metadata table exists and drop if it does
            if db.has_table(DEFAULT_METADATA_TABLE):
                db.drop_table(DEFAULT_METADATA_TABLE)

            # Setup metadata manager
            data_manager = DataManager(database=db)
            mdm = MetadataManager(data_manager)

            # Add metadata
            metadata = [{
                'title': 'Testing Data',
                'description': 'Data used for testing',
                'tags': 'test exp auto',
                'source': 'Automatically generated from Python',
                'created_by': 'msdss',
                'created_at': datetime.now(),
            }]
            mdm.create('test_data', metadata)
            before_update = mdm.get('test_data')

            # Update metadata
            mdm.updated_at('test_data', datetime.now())
            after_update = mdm.get('test_data')

            # Print results
            print('before_update:\\n')
            pprint(before_update)
            print('\\nafter_update:\\n')
            pprint(after_update)
        """
        data = {}
        data[self.updated_column] = dt if dt else datetime.now()
        self.update(name, data)


from copy import deepcopy
from datetime import datetime
from fastapi import APIRouter, Body, Depends, Query
from typing import Any, Dict, List, Literal, Optional

from .managers import *
from .models import *
from .tools import *

async def _no_current_user():
    return None

def get_data_router(
    users_api=None,
    database=None,
    route_settings=DEFAULT_DATA_ROUTE_SETTINGS,
    prefix='/data',
    tags=['data'],
    *args, **kwargs):
    """
    Get a data router.
    
    Parameters
    ----------
    users_api : :class:`msdss_users_api:msdss_users_api.core.UsersAPI` or None
        Users API object to enable user authentication for data routes.
        If ``None``, user authentication will not be used for data routes.
    database : :class:`msdss_base_database:msdss_base_database.core.Database` or None
        A :class:`msdss_base_database:msdss_base_database.core.Database` object for managing data. If ``None``, a default database will be used.
    route_settings : dict
        Dictionary of settings for the data routes. Each route consists of the following keys:

        * ``path``: resource path for the route
        * ``tags``: tags for open api spec
        * ``_enable`` (bool): Whether this route should be included or not
        * ``_restricted_tables`` (list(str)): List of table names not accessible by this route
        * ``_get_user`` (dict or None): Additional arguments passed to the :meth:`msdss_users_api.msdss_users_api.core.UsersAPI.get_current_user` function for the route - only applies if parameter ``users_api`` is not ``None`` and this settings is not ``None``, otherwise no user authentication will be added for this route
        * ``**kwargs``: Additional arguments passed to :meth:`fastapi:fastapi.FastAPI.get` for the id route

        The default settings are:

        .. jupyter-execute::
            :hide-code:

            from msdss_data_api.defaults import DEFAULT_DATA_ROUTE_SETTINGS
            from pprint import pprint
            pprint(DEFAULT_DATA_ROUTE_SETTINGS)

        Any unspecified settings will be replaced by their defaults.
    prefix : str
        Prefix path to all routes belonging to this router.
    tags : list(str)
        Tags for all routes in this router.
    *args, **kwargs
        Additional arguments to accept any extra parameters passed to :class:`fastapi:fastapi.routing.APIRouter`.
    
    Returns
    -------
    :class:`fastapi:fastapi.routing.APIRouter`
        A router object used for data routes. See `FastAPI bigger apps <https://fastapi.tiangolo.com/tutorial/bigger-applications/>`_

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------
    .. jupyter-execute::

        from msdss_base_database import Database
        from msdss_base_api import API
        from msdss_users_api import UsersAPI
        from msdss_data_api.routers import get_data_router

        # Create database object
        database = Database(
            driver='postgresql',
            user='msdss',
            password='msdss123',
            host='localhost',
            port='5432',
            database='msdss'
        )
        
        # Create an app
        app = API()

        # Add the data router
        router = get_data_router(database=database)
        app.add_router(router)

        # Add the data router with users
        # CHANGE SECRETS TO STRONG PHRASES
        app = API()
        users_api = UsersAPI(
            'cookie-secret',
            'jwt-secret',
            'reset-secret',
            'verification-secret',
            database=database
        )
        router = get_data_router(users_api, database=database)
        app.add_router(router)
        
        # Host app at https://localhost:8000
        # Try it at https://localhost:8000/docs
        # app.start()
    """ 
    database = database if database else Database()

    # (get_data_router_defaults) Merge defaults and user params 
    get_user = {}
    settings = deepcopy(DEFAULT_DATA_ROUTE_SETTINGS)
    for k in settings:
        if k in route_settings:
            settings[k].update(route_settings[k])

    # (get_data_router_apply) Apply settings to obtain dependencies
    get_user = {}
    get_data_manager = {}
    enable = {}
    for k, v in settings.items():
        get_user[k] = users_api.get_current_user(**v['_get_user']) if users_api and '_get_user' in v else _no_current_user
        del v['_get_user']
        get_data_manager[k] = create_data_manager_func(database=database, restricted_tables=v.pop('_restricted_tables'))
        enable[k] = v.pop('_enable')

    # (get_data_router_metamanager) Create metadata manager func
    get_metadata_manager = create_metadata_manager_func(database=database)

    # (get_data_router_create) Create api router for data routes
    out = APIRouter(prefix=prefix, tags=tags, *args, **kwargs)

    # (get_data_router_columns) Add columns route to data router
    if enable['columns']:
        @out.get(**settings['columns'])
        async def get_columns(
            dataset: str = Query(..., description='Name of the dataset'),
            data_manager = Depends(get_data_manager['query']),
            user = Depends(get_user['columns'])
        ):
            response = data_manager.get_columns(dataset)
            return response

    # (get_data_router_create) Add create route to data router
    if enable['create']:
        @out.post(**settings['create'])
        async def create_data(
            dataset: str = Query(..., description='Name of the dataset to create - the request body is used to upload JSON data under the "data" key in the form of "[{col: val, col2: val2, ...}, {col: val, col2: val2, ...}]", where each key represents a column and its corresponding value. Objects in this list should have the same keys.'),
            body: DataCreate = Body(
                ...,
                example={
                    'title': 'Title for Dataset',
                    'description': 'Description for dataset...',
                    'source': 'Data source for dataset',
                    'data': [
                        {'col_one': 1, 'col_two': 'a'},
                        {'col_one': 2, 'col_two': 'b'},
                        {'col_one': 3, 'col_two': 'c'}
                    ]
                }
            ),
            data_manager = Depends(get_data_manager['create']),
            metadata_manager = Depends(get_metadata_manager),
            user = Depends(get_user['create'])
        ):
            # (get_data_router_create_data) Get data
            body = body.dict()
            data = body.pop('data')

            # (get_data_router_create_metadata) Format metadata
            metadata = body
            metadata['dataset'] = dataset
            metadata['created_at'] = datetime.now()
            metadata['updated_at'] = datetime.now()

            # (get_data_router_create_users) Add user operations if available
            if user:
                metadata['created_by'] = user.email

            # (get_data_router_create_run) Create dataset and metadata
            data_manager.create(name=dataset, data=data)
            metadata_manager.create(name=dataset, data=metadata)

    # (get_data_router_delete) Add delete route to data router
    if enable['delete']:
        @out.delete(**settings['delete'])
        async def delete_data(
            dataset: str = Query(..., description='Name of the dataset to delete data from'),
            where: Optional[List[str]] = Query(None, description='Where statements to filter data to remove in the form of "column operator value" (e.g. "var < 3") - valid operators are: =, !=, >, >=, >, <, <=, !=, LIKE, ILIKE, NOTLIKE, NOTILIKE, CONTAINS, STARTSWITH, ENDSWITH'),
            where_boolean: Literal['AND', 'OR'] = Query('AND', alias='where-boolean', description='Either "AND" or "OR" to combine where statements'),
            delete_all: Optional[bool] = Query(False, description='Whether to remove the entire dataset or not'),
            data_manager = Depends(get_data_manager['delete']),
            metadata_manager = Depends(get_metadata_manager),
            user = Depends(get_user['delete'])
        ):
            data_manager.delete(name=dataset, where=where, where_boolean=where_boolean, delete_all=delete_all)
            metadata_manager.updated_at(name=dataset)
            if delete_all:
                metadata_manager.delete(name=dataset)

    # (get_data_router_id) Add id route to data router
    if enable['id']:
        @out.get(**settings['id'])
        async def get_data_by_id(
            dataset: str = Query(..., description='Name of the dataset'),
            id: str = Query(..., description='Identifier value to retrieve a specific document in the dataset'),
            id_column: Optional[str] =  Query('id', description='Identifier column name for the dataset'),
            data_manager = Depends(get_data_manager['id']),
            user = Depends(get_user['id'])
        ):
            where = [f'{id_column} = {id}']
            response = data_manager.get(name=dataset, where=where)
            return response

    # (get_data_router_insert) Add insert route to data router
    if enable['insert']:
        @out.put(**settings['insert'])
        async def insert_data(
            dataset: str = Query(..., description='Name of the dataset to insert - the request body is used to upload JSON data in the form of "[{key: value, key2: value2, ... }, {key: value, key2: value2, ...}]" where each key is a column name'),
            data: List[Dict[str, Any]] = Body(...),
            data_manager = Depends(get_data_manager['insert']),
            metadata_manager = Depends(get_metadata_manager),
            user = Depends(get_user['insert'])
        ):
            data_manager.insert(name=dataset, data=data)
            metadata_manager.updated_at(dataset)

    # (get_data_router_metadata) Add metadata route to data router
    if enable['metadata']:
        @out.get(**settings['metadata'])
        async def get_metadata(
            dataset: str = Query(..., description='Name of the dataset to get metadata for'),
            metadata_manager = Depends(get_metadata_manager),
            user = Depends(get_user['metadata'])
        ):
            response = metadata_manager.get(name=dataset)
            return response

    # (get_data_router_metadata) Add metadata route to data router
    if enable['metadata_update']:
        @out.put(**settings['metadata_update'])
        async def update_metadata(
            dataset: str = Query(..., description='Name of the dataset to update metadata for. Upload user and creation/update times can not be updated.'),
            body: MetadataUpdate = Body(
                ...,
                example={
                    'title': 'New Title to Replace Existing',
                    'description': 'New description to replace existing...',
                    'source': 'New data source to replace existing'
                }
            ),
            metadata_manager = Depends(get_metadata_manager),
            user = Depends(get_user['metadata_update'])
        ):
            response = metadata_manager.update(name=dataset, data=body.dict())
            return response

    # (get_data_router_query) Add query route to data router
    if enable['query']:
        @out.get(**settings['query'])
        async def query_data(
            dataset: str = Query(..., description='Name of the dataset to query'),
            select: Optional[List[str]] = Query('*', description='columns to include - "*" means all columns and "None" means to omit selection (useful for aggregate queries)'),
            where: Optional[List[str]] = Query(None, description='Where statements to filter data in the form of "column operator value" (e.g. "var < 3") - valid operators are: =, !=, >, >=, >, <, <=, !=, LIKE, ILIKE, NOTLIKE, NOTILIKE, CONTAINS, STARTSWITH, ENDSWITH'),
            group_by: Optional[List[str]] = Query(None, alias='group-by', description='column names to group by - should be used with aggregate and aggregate_func parameters'),
            aggregate: Optional[List[str]] = Query(None, description='column names to aggregate with the same order as the aggregate_func parameter'),
            aggregate_func: Optional[List[str]] = Query(None, alias='aggregate-func', description='Aggregate functions in the same order as the aggregate parameter'),
            order_by: Optional[List[str]] = Query(None, alias='order-by', description='column names to order by in the same order as parameter order_by_sort'),
            order_by_sort: Optional[List[Literal['asc', 'desc']]] = Query(None, alias='order-by-sort', description='Either "asc" for ascending or "desc" for descending order in the same order as parameter order_by'),
            limit: Optional[int] = Query(None, description='Number of items to return'),
            offset: Optional[int] = Query(None, description='Number of items to skip'),
            where_boolean: Literal['AND', 'OR'] = Query('AND', alias='where-boolean', description='Either "AND" or "OR" to combine where statements'),
            data_manager = Depends(get_data_manager['query']),
            user = Depends(get_user['query'])
        ):
            select = None if select[0] == 'None' else select
            response = data_manager.get(
                name=dataset,
                select=select,
                where=where,
                group_by=group_by,
                aggregate=aggregate,
                aggregate_func=aggregate_func,
                order_by=order_by,
                order_by_sort=order_by_sort,
                limit=limit,
                offset=offset,
                where_boolean=where_boolean
            )
            return response

    # (get_data_router_rows) Add rows route to data router
    if enable['rows']:
        @out.get(**settings['rows'])
        async def get_rows(
            dataset: str = Query(..., description='Name of the dataset'),
            data_manager = Depends(get_data_manager['query']),
            user = Depends(get_user['rows'])
        ):
            response = data_manager.get_rows(dataset)
            return response
    
    # (get_data_router_search) Add search route to data router
    if enable['search']:
        @out.get(**settings['search'])
        async def search_data(
            select: Optional[List[str]] = Query('*', description='columns to include in search - "*" means all columns and "None" means to omit selection (useful for aggregate queries).'),
            where: Optional[List[str]] = Query(None, description='Where statements to filter data in the form of "column operator value" (e.g. "dataset = test_data") - valid operators are: =, !=, >, >=, >, <, <=, !=, LIKE, ILIKE, NOTLIKE, NOTILIKE, CONTAINS, STARTSWITH, ENDSWITH'),
            order_by: Optional[List[str]] = Query(None, alias='order-by', description='column names to order by in the same order as parameter order_by_sort'),
            order_by_sort: Optional[List[Literal['asc', 'desc']]] = Query(None, alias='order-by-sort', description='Either "asc" for ascending or "desc" for descending order in the same order as parameter order_by'),
            limit: Optional[int] = Query(None, description='Number of items to return'),
            offset: Optional[int] = Query(None, description='Number of items to skip'),
            where_boolean: Literal['AND', 'OR'] = Query('AND', alias='where-boolean', description='Either "AND" or "OR" to combine where statements'),
            metadata_manager = Depends(get_metadata_manager),
            user = Depends(get_user['search'])
        ):
            select = None if select[0] == 'None' else select
            response = metadata_manager.search(
                select=select,
                where=where,
                order_by=order_by,
                order_by_sort=order_by_sort,
                limit=limit,
                offset=offset,
                where_boolean=where_boolean
            )
            return response
    
    # (get_data_router_update) Add update route to data router
    if enable['update']:
        @out.put(**settings['update'])
        async def update_data(
            dataset: str = Query(..., description='Name of the dataset to update - the request body is used to upload JSON data in the form of "{key: value, key2: value2, ... }" where each key is a column name and each value is the new value to use (matching the where parameter)'),
            body: Dict[str, Any] = Body(
                ...,
                example={'col_one': 1, 'col_two': 'a'}
            ),
            where: List[str] = Query(..., description='Where statements to filter data to update in the form of "column operator value" (e.g. "var < 3") - valid operators are: =, !=, >, >=, >, <, <=, !=, LIKE, ILIKE, NOTLIKE, NOTILIKE, CONTAINS, STARTSWITH, ENDSWITH'),
            data_manager = Depends(get_data_manager['update']),
            metadata_manager = Depends(get_metadata_manager),
            user = Depends(get_user['update'])
        ):
            data_manager.update(name=dataset, data=body, where=where)
            metadata_manager.updated_at(dataset)
    return out
